//Similar to extractFile.ts, but instead of receiving and returning an ArrayBuffer, works with Node.js streams.

import * as stream from 'stream';
import * as zlib from 'zlib';
import { ISsnFileEntry } from '../interfaces/ISsnFileEntry';
import decryptStream from './streams/decryptStream';
import waitReadableLength from './streams/waitReadableLength';

/** Extracts the file with the given metadata from the stream.
 * The stream must already start at the .zip's local file header
 * and must transparently span across multiple disks if necessary.
 */
export default async function extractFileStream(file: ISsnFileEntry, inputStream: stream.Readable, skipDecompression: boolean = false): Promise<stream.Readable> {
  let curStream = inputStream;

  //pipe into decryption if file is encrypted
  if (file.decryptionKeys !== undefined) {
    const decryptTransform = decryptStream(file.decryptionKeys);
    curStream = curStream.pipe(decryptTransform);
    //skip encryption header
    const encryptionHeader = curStream.read(12);
    if (encryptionHeader === null) {
      //need to wait until data is ready for reading
      await waitReadableLength(curStream, 12);
      curStream.read(12);
    }
  }

  if (!skipDecompression) {
    //pipe into decompression
    const decompressTransform = zlib.createInflateRaw();
    decompressTransform.on('error', (error) => {
      //TODO: need to throw error sync, not async
      throw new Error(`Error during decompression of "${file.name}": ${error.message}`);
    });
    curStream = curStream.pipe(decompressTransform);
  }

  return curStream;
}
