import downloadUrlContents from '../cdn/downloadUrlContents';
import getUrlContents from '../cdn/getUrlContents';
import { Product } from '../interfaces/ISettings';
import { SsnDiffType } from '../interfaces/ISsnFileEntry';
import extractFileStream from './extractFileStream';
import getSolidpkg from './getSolidpkg';
import readSsnFile from './reader/readSsnFile';
import getFileFromDisks from './streams/getFileFromDisks';
import streamToArrayBuffer from './streams/streamToArrayBuffer';
import verifyPatch from './verify/verifyPatch';

export default async function getPatch(product: Product, from: number, to: number) {
  const solidpkg = await getSolidpkg(product, from, to);
  console.debug(solidpkg.files);

  function createUrlObject({ name, length }: {name: string, length: number}) {
    return { host: 'cdn-patch.swtor.com', path: `/patch/${product}/${product}_${from}to${to}/${name}`, size: length };
  }

  //start download, making sure that .zip file downloads first
  const indexOfLastFile = solidpkg.files.length - 1;
  const zipFile = getUrlContents(createUrlObject(solidpkg.files[indexOfLastFile]));
  const diskFiles = solidpkg.files.slice(0, indexOfLastFile).map((file) => downloadUrlContents(createUrlObject(file)));

  //we can parse the file entries as soon as the .zip file is downloaded
  const fileEntries = readSsnFile(await zipFile);
  console.debug(fileEntries);

  //Verify file entries
  verifyPatch(fileEntries, product, from);

  //Then we need to wait for disks to finish download before we can extract individual files
  //TODO: we can optimize this to already extract some files as soon as their relevant parts are downloaded
  const diskFilenames = await Promise.all(diskFiles);
  //const dvArray = bufferArray.map((buffer) => new DataView(buffer));

  //TODO: Verify that downloaded files match the hash in `solidpkg.pieces`

  //Extract newly added files
  fileEntries.filter((file) => file.diffType === SsnDiffType.NewFile).forEach(async (file) => {
    try {
      const fileStream = await getFileFromDisks(diskFilenames, { diskStart: file.diskNumberStart, offset: file.offset, length: file.compressedSize });
      const fileContents = await extractFileStream(file, fileStream, true);
      console.debug(file.name, file.compressedSize, await streamToArrayBuffer(fileContents));
      //TODO: need to write to disk
    } catch (error) {
      console.error(`Could not extract file "${file.name}"`, error);
    }
  });

  //Extract changed files
  fileEntries.filter((file) => file.diffType === SsnDiffType.Changed).forEach(async (file) => {
    try {
      const fileStream = await getFileFromDisks(diskFilenames, { diskStart: file.diskNumberStart, offset: file.offset, length: file.compressedSize });
      const fileContents = await extractFileStream(file, fileStream, true);
      console.debug(file.name, file.compressedSize, await streamToArrayBuffer(fileContents));
      //TODO: need to apply diffing, then write to disk
    } catch (error) {
      console.error(`Could not extract file "${file.name}"`, error);
    }
  });

  //Need to delete deleted files
  fileEntries.filter((file) => file.diffType === SsnDiffType.Deleted).forEach((file) => {
    //TODO: need to delete file
  });
}
