import * as fs from 'fs';
import * as stream from 'stream';
import readLocalFileHeader from './readLocalFileHeader';

interface IGetFileFromDisksOptions {
  /** Number of the disk where the local file header starts */
  diskStart: number;
  /** Offset into the start of the disk where the local file header starts. */
  offset: number;
  /** Length of the stored file (compressed size + optional 12 byte encryption header), but excluding the length of the local file header. */
  length: number;
}

function createFileStream(disks: string[], index: number, offset: number, length: number = Infinity): fs.ReadStream {
  return fs.createReadStream(disks[index], { start: offset, end: offset + length - 1 });
}

/** Takes a list of ReadableStreams (the disks), as well as the offset and length, and returns a stream for just one file. */
async function getConcatenatedStream(disks: string[], { diskStart, offset, length }: IGetFileFromDisksOptions): Promise<stream.Readable> {
  let curDiskIndex = diskStart;
  let curDisk: fs.ReadStream = createFileStream(disks, diskStart, offset);
  let totalRead = 0;

  //Create new stream that concatenates disks until storedSize is reached, then ends the stream.
  const outputStream = new stream.PassThrough();

  const onData = (chunk: Buffer) => {
    totalRead += chunk.length;
    //If we've reached the end, we can stop reading after this chunk
    const readTooManyBytes = totalRead - length;
    if (readTooManyBytes >= 0) {
      //We can still write the whole chunk before ending the stream
      if (readTooManyBytes === 0) {
        outputStream.end(chunk);
        curDisk.close();
      } else {
        //We must shorten the chunk, write the shortened chunk and then end the stream
        const shortenedLength = chunk.length - readTooManyBytes;
        const shortenedChunk = Buffer.alloc(shortenedLength, chunk);
        outputStream.end(shortenedChunk);
        curDisk.close();
        curDisk.off('data', onData);
      }
    } else {
      //Nowhere near the end, so just write normally
      outputStream.write(chunk);
    }
  };
  const onError = (error: any) => {
    console.error(error);
  };
  const onEnd = () => {
    curDiskIndex += 1;
    //End if we are at end of file or end of disks
    if (curDiskIndex >= disks.length || totalRead >= length) {
      outputStream.end();
    } else {
      curDisk = createFileStream(disks, curDiskIndex, 0, length - totalRead);
      //set up new listeners for data and end
      curDisk.on('data', onData);
      curDisk.on('end', onEnd);
      curDisk.on('error', onError);
    }
  };

  curDisk.on('data', onData);
  curDisk.on('end', onEnd);
  curDisk.on('error', onError);

  return outputStream;
}

export default async function getFileFromDisks(disks: string[], { diskStart, offset, length }: IGetFileFromDisksOptions): Promise<stream.Readable> {
  //read local file header
  const headerStream = await getConcatenatedStream(disks, { diskStart, offset, length: 30 });
  const localFileHeaderLength = await readLocalFileHeader(headerStream);
  headerStream.destroy(); //TODO: is this the best way to close/destroy the stream?

  //read actual file
  return getConcatenatedStream(disks, { diskStart, offset: offset + localFileHeaderLength, length });
}
